# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Auxillary expectation functions
===============================

**Module name:** :mod:`pennylane_sf.expectations`

.. currentmodule:: pennylane_sf.expectations

Contains auxillary functions which convert from PennyLane-style expectations,
to the corresponding state methods in Strawberry Fields.

.. autosummary::
    identity
    mean_photon
    number_state
    homodyne
    poly_xp


Code details
~~~~~~~~~~~~
"""
import numpy as np

from strawberryfields.backends.states import BaseFockState
from strawberryfields.backends.gaussianbackend.states import GaussianState

import pennylane.expval


def identity(state, wires, params):
    """Computes the expectation value of the ``qml.expval.Identity``
    observable in Strawberry Fields, corresponding to the trace.

    Args:
        state (strawberryfields.backends.states.BaseState): the quantum state
        wires (Sequence[int]): the measured mode
        params (Sequence): sequence of parameters (not used)

    Returns:
        float, float: trace and its variance
    """
    # pylint: disable=unused-argument
    if isinstance(state, GaussianState):
        # Gaussian state representation will always have trace of 1
        return 1, 0

    N = state.num_modes
    D = state.cutoff_dim

    if N == len(wires):
        # trace of the entire system
        tr = state.trace()
        return tr, tr - tr**2

    # get the reduced density matrix
    N = len(wires)
    dm = state.reduced_dm(modes=wires)

    # construct the standard 2D density matrix, and take the trace
    new_ax = np.arange(2*N).reshape([N, 2]).T.flatten()
    tr = np.trace(dm.transpose(new_ax).reshape([D**N, D**N])).real

    return tr, tr - tr**2


def mean_photon(state, wires, params):
    """Computes the expectation value of the ``qml.expval.MeanPhoton``
    observable in Strawberry Fields.

    Args:
        state (strawberryfields.backends.states.BaseState): the quantum state
        wires (Sequence[int]): the measured mode
        params (Sequence): sequence of parameters (not used)

    Returns:
        float, float: mean photon number and its variance
    """
    # pylint: disable=unused-argument
    return state.mean_photon(wires[0])


def number_state(state, wires, params):
    """Computes the expectation value of the ``qml.expval.NumberState``
    observable in Strawberry Fields.

    Args:
        state (strawberryfields.backends.states.BaseState): the quantum state
        wires (Sequence[int]): the measured mode
        params (Sequence): sequence of parameters

    Returns:
        float, float: Fock state probability and its variance
    """
    # pylint: disable=unused-argument
    n = params[0]
    N = state.num_modes

    if N == len(wires):
        # expectation value of the entire system
        ex = state.fock_prob(n)
        return ex, ex - ex**2

    # otherwise, we must trace out remaining systems.
    if isinstance(state, BaseFockState):
        # fock state
        dm = state.reduced_dm(modes=wires)
        ex = dm[tuple([n[i//2] for i in range(len(n)*2)])].real

    elif isinstance(state, GaussianState):
        # Reduced Gaussian state
        mu, cov = state.reduced_gaussian(modes=wires)

        # calculate reduced Q
        Q = state._gmode.qmat() # pylint: disable=protected-access
        ind = np.concatenate([np.array(wires), N+np.array(wires)])
        rows = ind.reshape((-1, 1))
        cols = ind.reshape((1, -1))
        Q = Q[rows, cols]

        # calculate reduced Amat
        M = Q.shape[0]//2
        assert M == len(wires)

        I = np.identity(M)
        O = np.zeros_like(I)
        X = np.block([[O, I], [I, O]])
        A = X @ (np.identity(2*M)-np.linalg.inv(Q))

        # create reduced Gaussian state
        new_state = GaussianState((mu, cov), M, Q, A, hbar=state.hbar)
        ex = new_state.fock_prob(n)

    var = ex - ex**2
    return ex, var


def homodyne(phi=None):
    """Function factory that returns the ``qml.expval.Homodyne`` expectation
    function for Strawberry Fields.

    ``homodyne(phi)`` returns a function

    .. code-block:: python

        homodyne_expectation(state, wires, phi)

    that is used to determine the homodyne expectation value of a wire within a
    Strawberry Fields state object, measured along a particular phase-space
    angle ``phi``.

    Note that:

    * If ``phi`` is not None, the returned function will be hardcoded to return the
      homodyne expectation value at angle ``phi`` in the phase space.

    * If ``phi`` the value of ``phi`` must be set when calling the returned function.

    Args:
        phi (float): the default phase-space axis to perform the homodyne measurement on

    Returns:
        function: a function that accepts a SF state, the wire to measure,
        and phase space angle phi, and returns the quadrature expectation
        value and variance
    """
    if phi is not None:
        return lambda state, wires, params: state.quad_expectation(wires[0], phi)

    return lambda state, wires, params: state.quad_expectation(wires[0], *params)


def poly_xp(state, wires, params):
    r"""Computes the expectation value of an observable that is a second-order
    polynomial in :math:`\{\hat{x}_i, \hat{p}_i\}_i`.

    Args:
        state (strawberryfields.backends.states.BaseState): the quantum state
        wires (Sequence[int]): measured modes
        params (Sequence[array]): Q is a matrix or vector of coefficients
            using the :math:`(\I, \x_1,\p_1, \x_2,\p_2, \dots)` ordering

    Returns:
        float, float: expectation value, variance
    """
    Q = params[0]

    # HACK, we need access to the Poly instance in order to expand the matrix!
    op = pennylane.expval.PolyXP(Q, wires=wires, do_queue=False)
    Q = op.heisenberg_obs(state.num_modes)

    if Q.ndim == 1:
        d = np.r_[Q[1::2], Q[2::2]]
        return state.poly_quad_expectation(None, d, Q[0])

    # convert to the (I, x1,x2,..., p1,p2...) ordering
    M = np.vstack((Q[0:1, :], Q[1::2, :], Q[2::2, :]))
    M = np.hstack((M[:, 0:1], M[:, 1::2], M[:, 2::2]))
    d1 = M[1:, 0]
    d2 = M[0, 1:]
    return state.poly_quad_expectation(M[1:, 1:], d1+d2, M[0, 0])



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Strawberry Fields Gaussian device
=================================

**Module name:** :mod:`pennylane_sf.gaussian`

.. currentmodule:: pennylane_sf.gaussian

The Strawberry Fields Gaussian plugin implements all the :class:`~pennylane.device.Device` methods,
and provides a Gaussian simulation of a continuous-variable quantum circuit.

Classes
-------

.. autosummary::
   StrawberryFieldsGaussian


Code details
~~~~~~~~~~~~
"""

import numpy as np

#import state preparations
from strawberryfields.ops import (Coherent, DisplacedSqueezed,
                                  Squeezed, Thermal, Gaussian)
# import gates
from strawberryfields.ops import (BSgate, CXgate, CZgate, Dgate,
                                  Pgate, Rgate, S2gate, Sgate)

from .expectations import (identity, mean_photon, homodyne, number_state, poly_xp)
from .simulator import StrawberryFieldsSimulator


class StrawberryFieldsGaussian(StrawberryFieldsSimulator):
    """StrawberryFields Gaussian device for PennyLane.
    """
    name = 'Strawberry Fields Gaussian PennyLane plugin'
    short_name = 'strawberryfields.gaussian'

    _operation_map = {
        'CoherentState': Coherent,
        'DisplacedSqueezedState': DisplacedSqueezed,
        'SqueezedState': Squeezed,
        'ThermalState': Thermal,
        'GaussianState': Gaussian,
        'Beamsplitter': BSgate,
        'ControlledAddition': CXgate,
        'ControlledPhase': CZgate,
        'Displacement': Dgate,
        'QuadraticPhase': Pgate,
        'Rotation': Rgate,
        'TwoModeSqueezing': S2gate,
        'Squeezing': Sgate
    }

    _expectation_map = {
        'MeanPhoton': mean_photon,
        'X': homodyne(0),
        'P': homodyne(np.pi/2),
        'Homodyne': homodyne(),
        'PolyXP': poly_xp,
        'NumberState': number_state,
        'Identity': identity
    }

    _circuits = {}

    def pre_expval(self):
        self.state = self.eng.run('gaussian')



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Strawberry Fields Fock device
=============================

**Module name:** :mod:`pennylane_sf.fock`

.. currentmodule:: pennylane_sf.fock

The Strawberry Fields Fock plugin implements all the :class:`~pennylane.device.Device` methods,
and provides a Fock-space simulation of a continuous-variable quantum circuit.

Classes
-------

.. autosummary::
   StrawberryFieldsFock

----
"""

import numpy as np

#import state preparations
from strawberryfields.ops import (Catstate, Coherent, DensityMatrix, DisplacedSqueezed,
                                  Fock, Ket, Squeezed, Thermal, Gaussian)
# import gates
from strawberryfields.ops import (BSgate, CKgate, CXgate, CZgate, Dgate,
                                  Kgate, Pgate, Rgate, S2gate, Sgate, Vgate)

from .expectations import (identity, mean_photon, homodyne, number_state, poly_xp)
from .simulator import StrawberryFieldsSimulator


class StrawberryFieldsFock(StrawberryFieldsSimulator):
    r"""StrawberryFields Fock device for PennyLane.

    Args:
        wires (int): the number of modes to initialize the device in
        shots (int): Number of circuit evaluations/random samples
            used to estimate expectation values of observables.
            For simulator devices, 0 means the exact EV is returned.
        cutoff_dim (int): Fock-space truncation dimension
        hbar (float): the convention chosen in the canonical commutation
            relation :math:`[x, p] = i \hbar`
    """
    name = 'Strawberry Fields Fock PennyLane plugin'
    short_name = 'strawberryfields.fock'

    _operation_map = {
        'CatState': Catstate,
        'CoherentState': Coherent,
        'FockDensityMatrix': DensityMatrix,
        'DisplacedSqueezedState': DisplacedSqueezed,
        'FockState': Fock,
        'FockStateVector': Ket,
        'SqueezedState': Squeezed,
        'ThermalState': Thermal,
        'GaussianState': Gaussian,
        'Beamsplitter': BSgate,
        'CrossKerr': CKgate,
        'ControlledAddition': CXgate,
        'ControlledPhase': CZgate,
        'Displacement': Dgate,
        'Kerr': Kgate,
        'QuadraticPhase': Pgate,
        'Rotation': Rgate,
        'TwoModeSqueezing': S2gate,
        'Squeezing': Sgate,
        'CubicPhase': Vgate
    }

    _expectation_map = {
        'MeanPhoton': mean_photon,
        'X': homodyne(0),
        'P': homodyne(np.pi/2),
        'Homodyne': homodyne(),
        'PolyXP': poly_xp,
        'NumberState': number_state,
        'Identity': identity
    }

    _circuits = {}

    def __init__(self, wires, *, cutoff_dim, shots=0, hbar=2):
        super().__init__(wires, shots=shots, hbar=hbar)
        self.cutoff = cutoff_dim

    def pre_expval(self):
        self.state = self.eng.run('fock', cutoff_dim=self.cutoff)



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#!/usr/bin/env python3
import sys
import os
from setuptools import setup

with open("pennylane_sf/_version.py") as f:
	version = f.readlines()[-1].split()[-1].strip("\"'")


requirements = [
    "strawberryfields>=0.9",
    "pennylane>=0.2"
]

info = {
    'name': 'PennyLane-SF',
    'version': version,
    'maintainer': 'Xanadu Inc.',
    'maintainer_email': 'nathan@xanadu.ai',
    'url': 'http://xanadu.ai',
    'license': 'Apache License 2.0',
    'packages': [
                    'pennylane_sf'
                ],
    'entry_points': {
        'pennylane.plugins': [
            'strawberryfields.fock = pennylane_sf:StrawberryFieldsFock',
            'strawberryfields.gaussian = pennylane_sf:StrawberryFieldsGaussian'
            ],
        },
    'description': 'Open source library for continuous-variable quantum computation',
    'long_description': open('README.rst').read(),
    'provides': ["pennylane_sf"],
    'install_requires': requirements,
    # 'extras_require': extra_requirements,
    'command_options': {
        'build_sphinx': {
            'version': ('setup.py', version),
            'release': ('setup.py', version)}}
}

classifiers = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: Apache Software License",
    "Natural Language :: English",
    "Operating System :: POSIX",
    "Operating System :: MacOS :: MacOS X",
    "Operating System :: POSIX :: Linux",
    "Operating System :: Microsoft :: Windows",
    "Programming Language :: Python",
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3 :: Only',
    "Topic :: Scientific/Engineering :: Physics"
]

setup(classifiers=classifiers, **(info))



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Base simulator class
====================

**Module name:** :mod:`pennylane_sf.simulator`

.. currentmodule:: pennylane_sf.simulator

A base class for constructing Strawberry Fields devices for PennyLane.
This class provides all the boilerplate for supporting PennyLane;
inheriting devices simply need to provide their engine run command
in :meth:`~.StrawberryFieldsSimulator.pre_expval`, as well as defining their ``_operation_map``
and ``_expectation_map``, mapping PennyLane operations to their
Strawberry Fields counterparts.

Classes
-------

.. autosummary::
   StrawberryFieldsSimulator

Code details
~~~~~~~~~~~~
"""
import abc

import numpy as np

from pennylane import Device
import strawberryfields as sf

from ._version import __version__


class StrawberryFieldsSimulator(Device):
    r"""Abstract StrawberryFields simulator device for PennyLane.

    Args:
        wires (int): the number of modes to initialize the device in
        shots (int): Number of circuit evaluations/random samples used
            to estimate expectation values of expectations.
            For simulator devices, 0 means the exact EV is returned.
        hbar (float): the convention chosen in the canonical commutation relation :math:`[x, p] = i \hbar`
    """
    name = 'Strawberry Fields Simulator PennyLane plugin'
    pennylane_requires = '0.2.0'
    version = __version__
    author = 'Josh Izaac'

    short_name = 'strawberryfields'
    _operation_map = {}
    _expectation_map = {}

    def __init__(self, wires, *, shots=0, hbar=2):
        super().__init__(wires, shots)
        self.hbar = hbar
        self.eng = None
        self.q = None
        self.state = None

    def execution_context(self):
        """Initialize the engine"""
        self.reset()
        self.eng, self.q = sf.Engine(self.num_wires, hbar=self.hbar)
        return self.eng

    def apply(self, operation, wires, par):
        """Apply a quantum operation.

        Args:
            operation (str): name of the operation
            wires (Sequence[int]): subsystems the operation is applied on
            par (tuple): parameters for the operation
        """
        # convert PennyLane parameter conventions to
        # Strawberry Fields conventions
        if operation == "DisplacedSqueezedState":
            sf_par = (par[0]*np.exp(par[1]*1j), par[2], par[3])
        elif operation == "CatState":
            sf_par = (par[0]*np.exp(par[1]*1j), par[2])
        else:
            sf_par = par

        op = self._operation_map[operation](*sf_par)
        op | [self.q[i] for i in wires] #pylint: disable=pointless-statement

    @abc.abstractmethod
    def pre_expval(self):
        """Run the engine"""
        raise NotImplementedError

    def expval(self, expectation, wires, par):
        """Evaluate an expectation.

        Args:
            expectation (str): name of the expectation
            wires (Sequence[int]): subsystems the expectation is evaluated on
            par (tuple): parameters for the expectation
        Returns:
            float: expectation value
        """
        ex, var = self._expectation_map[expectation](self.state, wires, par)

        if self.shots != 0:
            # estimate the expectation value
            # use central limit theorem, sample normal distribution once, only ok
            # if shots is large (see https://en.wikipedia.org/wiki/Berry%E2%80%93Esseen_theorem)
            ex = np.random.normal(ex, np.sqrt(var / self.shots))

        return ex

    def reset(self):
        """Reset the device"""
        if self.eng is not None:
            self.eng.reset()
            self.eng = None
        if self.state is not None:
            self.state = None
        if self.q is not None:
            self.q = None

    @property
    def operations(self):
        """Get the supported set of operations.

        Returns:
            set[str]: the set of PennyLane operation names the device supports
        """
        return set(self._operation_map.keys())

    @property
    def expectations(self):
        """Get the supported set of expectations.

        Returns:
            set[str]: the set of PennyLane expectation names the device supports
        """
        return set(self._expectation_map.keys())
